package src.mlp;

import src.mlp.activationFunctions.LinearFunction;
import src.mlp.activationFunctions.SigmoidFunction;
import src.mlp.statistics.ConfusionMatrix;

/*Classe que representa uma rede neural que ser utilizada como folha na rvore de reconhecimento.*/
public class Network {
	private double learningRate;	
	private boolean print = false;
	private double [] inputLayer;
	private double [] outputLayerDeltas;
	private double [] hiddenLayerDeltas;
	private int numberOfClasses;	
	private PerceptronLayer hiddenLayer;
	private PerceptronLayer outputLayer;
	private WeightLayer inputToHiddenWeight;
	private WeightLayer hiddenToOutputWeight;
	private SigmoidFunction sigmoid = new SigmoidFunction();
	private LinearFunction linear = new LinearFunction();
	private ConfusionMatrix confusionMatrix;
	
	
	public Network (NetworkConfigs netSize){
		int inputPerceptrons = netSize.getNumberOfInputPerceptrons();
		int hiddenPerceptrons = netSize.getNumberOfHiddenPerceptrons();
		int outputPerceptrons = netSize.getNumberOfOutputPerceptrons();
		numberOfClasses = outputPerceptrons;		
		hiddenLayer = new PerceptronLayer (hiddenPerceptrons, sigmoid);
		outputLayer = new PerceptronLayer (outputPerceptrons, linear);
		hiddenToOutputWeight = new WeightLayer(hiddenPerceptrons, outputPerceptrons);		
		inputToHiddenWeight = new WeightLayer (inputPerceptrons, hiddenPerceptrons);
		setConfusionMatrix(new ConfusionMatrix(numberOfClasses));
		this.learningRate = netSize.getLearningRate();
	}
	
	/*Classe que recebe o vetor esperado e o vetor de sada produzido pela rede 
	 * e retorna o erro usando a soma dos quadrados */
	public double calculateError (int desiredOutput ){
		int i ;
		double result = 0;
		//System.out.println("calculando erro para resultado esperado: "+desiredOutput);
		for(i=0; i < outputLayer.getPerceptron().length-1; i++){
			/*O resultado esperado  1 para a posio desiredOutput do vetor e zero para as demais.*/
			if (i == desiredOutput){				
				result = result + Math.pow(1 - outputLayer.getOutput(i),2);			
			}
			else
				result = result + Math.pow( - outputLayer.getOutput(i),2);			
		}		
		return result/2;
	}
	
	private void calculateHiddenLayerDeltas(){
		double sum;
		hiddenLayerDeltas = new double [hiddenLayer.getPerceptron().length];
		for(int j = 0;  j < hiddenLayer.getOutput().length; j++){
			sum = 0;
			for(int k = 0; k < outputLayer.getOutput().length-1; k++){
				sum = sum + (hiddenToOutputWeight.getWeight(j, k)* outputLayerDeltas[k]);
			}
			hiddenLayerDeltas[j] = hiddenLayer.getOutput(j)*(1-hiddenLayer.getOutput(j))*sum;					
		}		
	}
	
	private void calculateOutputLayerDeltas (int desiredOutput){
		outputLayerDeltas = new double[outputLayer.getPerceptron().length];
		for(int k= 0; k< outputLayer.getOutput().length-1; k++){
			if(k == desiredOutput){
				outputLayerDeltas[k] = 1 - outputLayer.getOutput(k);				
			}
			else
				outputLayerDeltas[k] = - outputLayer.getOutput(k);			
		}
	}
	
	
	private void updateHiddenToOutputWeights(){
		double newWeight;
		for (int j = 0; j < hiddenLayer.getPerceptron().length; j++  )
			for(int k =0 ; k < outputLayer.getPerceptron().length-1 ; k++){				
				newWeight = (hiddenToOutputWeight.getWeight(j, k)-(learningRate * outputLayerDeltas[k]* hiddenLayer.getOutput(j)));
				
				hiddenToOutputWeight.setWeight(j, k, newWeight);
			}	
	}
	private void updateInputToHiddenWeights(){
		double newWeight;
		int i,j;
		for( i = 0; i< inputLayer.length ; i++)
			for( j = 0; j< hiddenLayer.getPerceptron().length-1; j++){				
				newWeight = (inputToHiddenWeight.getWeight(i, j)- (learningRate * hiddenLayerDeltas[j] * inputLayer[i]));
				inputToHiddenWeight.setWeight(i, j, newWeight);				
			}		
	}
	
	public void updateWeights(int desiredOutput, double []output){
		calculateOutputLayerDeltas(desiredOutput);
		calculateHiddenLayerDeltas();		
		updateHiddenToOutputWeights();		
		updateInputToHiddenWeights();		
	}	
	
	private double[] initializeInputLayer(PatternSymbol ps){
		double[] array = new double[ps.getSymbol().length+1];
		for(int i =0;i<ps.getSymbol().length;i++)
			array[i]=ps.getSymbol()[i];
		array[ps.getSymbol().length]=1;
		return array;
	}		
	
	/*Cada camada da rede faz o processamento. Ao trmino  retornada a classe que teve o maior valor de sada.*/
	public int recognize (PatternSymbol ps){
		int i;		
		inputLayer = initializeInputLayer(ps);
		hiddenLayer.produceOutput(inputLayer, inputToHiddenWeight);
		outputLayer.produceOutput(hiddenLayer.getOutput(), hiddenToOutputWeight );
		double greatestValue = 0;
		int greatestValueIndex = 0;
		
		/*Encontra o valor mximo no vetor de sadas*/		
		for (i = 0 ; i<outputLayer.size()-1; i++){
			if(print)
				System.out.println("calculando maior valor de saida. i vale " +i+". valor  "+outputLayer.getOutput(i)  );
			if(outputLayer.getOutput(i) > greatestValue){
				greatestValue = outputLayer.getOutput(i);
				greatestValueIndex = i;
			}
		}		
		return greatestValueIndex;		
	}
	public double[] getOutputLayerOutput(){
		return outputLayer.getOutput();
	}

	public void setConfusionMatrix(ConfusionMatrix confusionMatrix) {
		this.confusionMatrix = confusionMatrix;
	}

	public ConfusionMatrix getConfusionMatrix() {
		return confusionMatrix;
	}	
}
